@@ -17,36 +17,13 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
-import sys, os, re
+import sys, os, re, math
 from owlready2 import *
-
-"""
-
-Not supported:
-
-ASK
-DESCRIBE
-LOAD, ADD, MOVE, COPY, CLEAR, DROP
-
-GRAPH, FROM, FROM NAMED
-
-CONSTRUCT
-
-SERVICE
-
-INSERT DATA, DELETE DATA, DELETE WHERE (use INSERT or DELETE instead)
-
-VALUES
-
-MINUS
-
-"""
-
 from owlready2.sparql.parser import *
 from owlready2.sparql.func   import register_python_function, FuncSupport
 
-_RE_AUTOMATIC_INDEX = re.compile(r"TABLE (.*?) AS (.*?) USING AUTOMATIC.*\((.*?)\)")
-_RE_NORMAL_INDEX    = re.compile(r"TABLE (.*?) AS (.*?) USING (COVERING )?INDEX (.*?) ")
+_RE_AUTOMATIC_INDEX = re.compile(r"([^ ]*?) USING AUTOMATIC.*\((.*?)\)")
+#_RE_NORMAL_INDEX    = re.compile(r"(.*?) AS (.*?) USING (COVERING )?INDEX (.*?) ")
 
 _DEPRIORIZE_SUBQUERIES_OPT = True
 
@@ -58,21 +35,33 @@ class Translator(object):
     self.base_iri                      = ""
     self.current_anonynous_var         = 0
     self.current_parameter             = 0
+    self.max_fixed_parameter           = 0
     self.main_query                    = None
     self.preliminary_selects           = []
     self.recursive_preliminary_selects = {}
     self.escape_mark                   = "@@@ESCAPE@@@"
+    self.next_table_id                 = 1
+    self.table_name_2_type             = {}
+    self.table_type_2_cols             = { "objs" : ["s", "p", "o"], "datas" : ["s", "p", "o", "d"], "quads" : ["s", "p", "o", "d"] , "one" : ["i"] }
     
     if not getattr(world.graph, "_has_sparql_func", False):
       register_python_function(world)
       world.graph._has_sparql_func = True
       
+  def make_translator(self):
+    translator = Translator(self.world, self.error_on_undefined_entities)
+    translator.prefixes = self.prefixes.copy()
+    translator.base_iri = self.base_iri
+    return translator
+  
   def parse(self, sparql):
     while self.escape_mark in sparql:
       self.escape_mark += "ç"
     CURRENT_TRANSLATOR.set(self)
     self.main_query = PARSER.parse(LEXER.lex(sparql))
+    return self.finalize()
     
+  def finalize(self):
     sql = ""
     if self.preliminary_selects:
       sql += "WITH "
@@ -93,7 +82,7 @@ class Translator(object):
       if not self.solution_modifier[3]: sql += " LIMIT -1" # SQLite requires a LIMIT clause before the OFFSET clause
       sql += " OFFSET %s" % self._to_sql(self.solution_modifier[4])
       
-    nb_parameter = self.current_parameter
+    nb_parameter = max(self.current_parameter, self.max_fixed_parameter)
     parameter_2_parameter_datatypes = {}
     parameter_datatypes = []
     if self.escape_mark in sql:
@@ -107,8 +96,11 @@ class Translator(object):
             parameter_datatypes.append(number - 1)
         return "?%s" % r
       sql = re.sub("%s[^ ]*" % self.escape_mark, sub, sql)
-
-    #sql = self.optimize_sql(sql)
+      
+    #if sql:
+    #  if   self.main_query.type == "select": nb_sql_parameter = nb_parameter + len(parameter_datatypes)
+    #  else:                                  nb_sql_parameter = len(self.main_query.select_param_indexes)
+    #  sql = self.optimize_sql(sql, nb_sql_parameter)
     
     if   self.main_query.type == "select":
       return PreparedSelectQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes)
@@ -118,6 +110,27 @@ class Translator(object):
       return PreparedModifyQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes, self.world.get_ontology(self.main_query.ontology_iri.value) if self.main_query.ontology_iri else None, self.parse_inserts_deletes(self.main_query.deletes, self.main_query.columns), self.parse_inserts_deletes(self.main_query.inserts, self.main_query.columns), select_param_indexes)
     
     
+  def optimize_sql(self, sql, nb_sql_parameter):
+    plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql, (1,) * nb_sql_parameter))
+                
+    has_automatic_index = False
+    for l in plan:
+      match = _RE_AUTOMATIC_INDEX.search(l[3])
+      if match:
+        table_name = match.group(1)
+        index_name = match.group(2)
+        table_type = self.table_name_2_type.get(table_name)
+        
+        if (table_type == "objs") or (table_type == "datas"):
+          if   index_name.startswith("s="): index_name = "index_%s_sp" % table_type
+          elif index_name.startswith("o="): index_name = "index_%s_op" % table_type
+          else: continue
+          
+          print("OPTIMIZE!!!", l, table_type, table_name, "=>", index_name)
+          table_def = "%s %s" % (table_type, table_name)
+          sql = sql.replace(table_def, "%s INDEXED BY %s" % (table_def, index_name), 1)
+    return sql
+    
   # def optimize_sql(self, sql):
   #   # Avoid Sqlite3 AUTOMATIC INDEX when a similar index can be used.
   #   plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
@@ -160,7 +173,7 @@ class Translator(object):
             triple.append(("vars", column.index))
             if len(triple) == 3: # in 'o' position => can be objs or datas!
               if column.index + 1 < len(columns):
-                next_column = columns[column.index]
+                next_column = columns[column.index + 1]
                 if next_column.name.endswith("d"):
                   triple.append(("vars", next_column.index))
         elif x.name == "PARAM":
@@ -179,6 +192,7 @@ class Translator(object):
             triple.append(("datas", d2))
             
       r.append(triple)
+
     return r
   
   def new_sql_query(self, name, block, selects = None, distinct = None, solution_modifier = None, preliminary = False, extra_binds = None, nested_inside = None, copy_vars = False):
@@ -196,15 +210,15 @@ class Translator(object):
       
     elif isinstance(block, UnionBlock):
       s = SQLCompoundQuery(name, nested_inside)
-      if selects is None: selects = block.get_ordered_vars()
+      if selects is None:
+        selects = block.get_ordered_vars()
         
     elif isinstance(block, FilterBlock):
       s = SQLNestedQuery(name)
       #if isinstance(block, ExistsBlock): s.extra_sql = "IS NOT NULL"
       #else:                              s.extra_sql = "IS NULL"
       s.exists = isinstance(block, ExistsBlock)
-      s.next_table_id = nested_inside.next_table_id
-      s.vars = nested_inside.vars
+      s.vars   = nested_inside.vars
       preliminary = False
       
     elif isinstance(block, NotExistsBlock):
@@ -246,7 +260,7 @@ class Translator(object):
       for bind in extra_binds: s.prepare_bind(bind)
       
       s.parse_selects(selects)
-      
+
       s.parse_triples(block)
       for bind in extra_binds: s.parse_bind(bind)
       
@@ -254,15 +268,11 @@ class Translator(object):
       
     elif isinstance(block, UnionBlock):
       for alternative in block:
-        #print("   ", s.parent.vars)
         query = self.new_sql_query(None, alternative, selects, distinct, None, False, extra_binds, nested_inside = s, copy_vars = False)
         s.append(query, "UNION")
       s.finalize_compounds()
       
       
-    if isinstance(block, FilterBlock):
-      nested_inside.next_table_id = s.next_table_id
-      
     if (not preliminary) and solution_modifier: self.solution_modifier = solution_modifier
     return s
 
@@ -313,7 +323,7 @@ class PreparedQuery(object):
     self.column_types        = column_types
     self.nb_parameter        = nb_parameter
     self.parameter_datatypes = parameter_datatypes
-  
+    
   def execute(self, params = ()):
     self.world._nb_sparql_call += 1
     sql_params = [self.world._to_rdf(param)[0] for param in params]
@@ -339,6 +349,26 @@ class PreparedSelectQuery(PreparedQuery):
           i += 2
       yield l2
       
+  def _execute_sql(self, params = ()):
+    for l in PreparedQuery.execute(self, params):
+      l2 = []
+      i = 0
+      while i < len(l):
+        #if self.column_types[i] == "objs":
+        if (self.column_types[i] == "objs") or (self.column_types[i] == "value"):
+          l2.append(l[i])
+          i += 1
+        else:
+          if l[i + 1] is None:
+            if l[i] is None: v = None
+            else:            v = self.world._to_python(l[i], None) or l[i]
+          else:
+            v = self.world._to_python(l[i], l[i + 1])
+          if isinstance(v, str): l2.append("'%s'" % v.replace("'", "\\'"))
+          else:                  l2.append(v)
+          i += 2
+      yield l2
+      
   def execute_flat(self, params = ()):
     for l in PreparedQuery.execute(self, params):
       i = 0
@@ -458,6 +488,24 @@ class PreparedSelectQuery(PreparedQuery):
 """
     return xml
   
+  def execute_as_sql(self, params = ()):
+    for l in PreparedQuery.execute(self, params):
+      l2 = []
+      i = 0
+      while i < len(l):
+        if self.column_types[i] == "objs":
+          if l[i] is None: l2.append(None)
+          else: l2.append(self.world._to_python(l[i], None) or l[i])
+          i += 1
+        else:
+          if l[i + 1] is None:
+            if l[i] is None: l2.append(None)
+            else:            l2.append(self.world._to_python(l[i], None) or l[i])
+          else:
+            l2.append(self.world._to_python(l[i], l[i + 1]))
+          i += 2
+      yield l2
+      
   
 class PreparedModifyQuery(PreparedQuery):
   def __init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes, ontology, deletes, inserts, select_param_indexes):
@@ -472,8 +520,10 @@ class PreparedModifyQuery(PreparedQuery):
   def execute(self, params = ()):
     nb_match = 0
     if self.sql: resultss = PreparedQuery.execute(self, [params[i] for i in self.select_param_indexes])
-    else:        resultss = [[]]
-    for results in list(resultss):
+    else:        resultss = [()]
+    
+    added_triples = []
+    for results in set(resultss):
       nb_match += 1
       
       for delete in self.deletes:
@@ -495,9 +545,10 @@ class PreparedModifyQuery(PreparedQuery):
           elif type == "param":         triple.append(self.world._to_rdf(params[value])[0])
           elif type == "paramdatatype": triple.append(self.world._to_rdf(params[value])[1])
           else:                         triple.append(value)
-        #print("ADD", triple)
-        self.world._add_triple_with_update(self.ontology, *triple)
+        #print("ADD", insert, triple)
+        added_triples.append(triple)
         
+    if added_triples: self.world._add_triples_with_update(self.ontology, added_triples)
     return nb_match
   
     
@@ -528,11 +579,9 @@ class Variable(object):
   def get_binding(self, query):
     if not self.bindings:
       #print("* Owlready2 * WARNING: variable without binding in SPARQL, use a suboptimal option", file = sys.stderr)
-      table = Table("any%s" % query.next_table_id, """(SELECT DISTINCT s AS o, NULL AS d FROM quads UNION SELECT DISTINCT o, d FROM quads)""")
-      table.query = query
-      query.tables.append(table)
-      query.name_2_table[table.name] = table
-      query.next_table_id += 1
+      table = Table(query, "any%s" % query.translator.next_table_id, """(SELECT DISTINCT s AS o, NULL AS d FROM quads UNION SELECT DISTINCT o, d FROM quads)""")
+      table.subquery = query
+      query.translator.next_table_id += 1
       self.bindings.append("%s.o" % table.name)
     i = 0
     for binding in self.bindings:
@@ -541,20 +590,24 @@ class Variable(object):
   
   def update_type(self, type):
     if   self.type == "quads": self.type = type
-    elif (type != self.type) and (type != "quads") :
-      raise ValueError("Variable %s cannot be both %s and %s!" % (self.name, self.type, type))
+    elif (type != self.type) and (type != "quads"):
+      raise ValueError("Variable %s cannot be both %s and %s!\n\n(NB if you are querying entities that are objects of a rdfs:label or comment relations, please add the following code to prevent Owlready from assuming that label and comment are data and not entities:\n    import owlready2.sparql.parser\n    owlready2.sparql.parser._DATA_PROPS = set()\n)" % (self.name, self.type, type))
     
 class Table(object):
-  def __init__(self, name, type = "quads", index = None, join = ",", join_conditions = None):
+  def __init__(self, query, name, type = "quads", index = None, join = ",", join_conditions = None):
     self.name            = name
     self.type            = type
     self.index           = index
     self.join            = join
     self.join_conditions = join_conditions or []
-    self.query           = None
-    
+    self.subquery        = None
+    if query:
+      query.tables.append(self)
+      query.name_2_table[name] = self
+      query.translator.table_name_2_type[name] = type
+      
   def __repr__(self): return "<Table '%s %s'>" % (self.type, self.name)
-    
+  
   def sql(self):
     return """%s %s%s%s""" % (self.type, self.name, self.index and (" INDEXED BY %s" % self.index) or "", self.join_conditions and (" ON (%s)" % " AND ". join(self.join_conditions)) or "")
   
@@ -567,7 +620,6 @@ class SQLQuery(FuncSupport):
     self.translator               = CURRENT_TRANSLATOR.get()
     self.distinct                 = False
     self.raw_selects              = None
-    self.next_table_id            = 1
     self.tables                   = []
     self.name_2_table             = {}
     self.columns                  = []
@@ -578,7 +630,6 @@ class SQLQuery(FuncSupport):
     self.extra_sql                = ""
     self.select_simple_union      = False
     self.optional                 = False
-    self.has_one                  = False
     
   def __repr__(self): return "<%s '%s'>" % (self.__class__.__name__, self.sql())
   
@@ -605,20 +656,30 @@ class SQLQuery(FuncSupport):
             else:      l[-1].append(column.binding)
         sql = """VALUES %s""" % ",".join("(%s)" % ",".join(str(k) for k in j) for j in l)
       else:
-        sql = """VALUES (%s)""" % (",".join(str(column.binding) for column in self.columns))
+        vars = []
+        for column in self.columns:
+          if isinstance(column.binding, str) and column.binding.startswith("IN "):
+            var = self.parse_var(column.var)
+            vars.append(var)
+        if vars:
+          if hasattr(vars[0], "static"):
+            sql = """VALUES %s""" % ",".join("(%s)" % (",".join(str(value) for value in values)) for values in zip(*[[i[0] for i in var.static] for var in vars]))
+          else:
+            sql = """SELECT %s FROM %s""" % (",".join("%s.col1_o" % var.in_select.name for var in vars), ",".join(var.in_select.name for var in vars))
+        else:
+          sql = """VALUES (%s)""" % (",".join(str(column.binding) for column in self.columns))
         
     if self.extra_sql: sql += " %s" % self.extra_sql
-    if self.preliminary: return """%s(%s) AS (%s)""" % (self.name, ", ".join(column.name for column in self.columns), sql)
+    if self.preliminary:
+      return """%s(%s) AS (%s)""" % (self.name, ", ".join(column.name for column in self.columns), sql)
     return sql
     
-  def set_column_names(self, names):
-    for column, name in zip(self.columns, names): column.name = name
-    
   def parse_distinct(self, distinct):
     if isinstance(distinct, rply.Token): self.distinct = distinct and (distinct.value.upper() == "DISTINCT")
     else:                                self.distinct = distinct
     
   def parse_var(self, x):
+    if isinstance(x, Variable):   return x
     if isinstance(x, rply.Token): name = x.value
     else:                         name = x
     var = self.vars.get(name)
@@ -657,32 +718,30 @@ class SQLQuery(FuncSupport):
   def parse_filter(self, filter):
     sql = self.parse_expression(filter.constraint)
     self.conditions.append(sql)
-    
+              
   def add_subquery(self, sub):
     if isinstance(sub, SQLNestedQuery):
       self.conditions.append(sub)
-      if _DEPRIORIZE_SUBQUERIES_OPT and not self.has_one:
-        self.tables.append(Table("one", "one"))
-        self.has_one = True
+      if _DEPRIORIZE_SUBQUERIES_OPT and not ("one" in self.name_2_table): Table(self, "one", "one")
     else:
-      table = Table("p%s" % self.next_table_id, sub.name)
-      table.query = sub
-      self.next_table_id += 1
-      self.tables.append(table)
-      self.name_2_table[table.name] = table
-      if sub.optional:
-        table.join = "LEFT JOIN"
-        conditions = table.join_conditions
+      if (not sub.optional) and (len(sub.columns) == 1): # Specific optimization with IN operator
+        self.create_in_conditions(self.conditions, sub.columns[0].var, sub)
       else:
-        conditions = self.conditions
-
-      for column in sub.columns:
-        var = self.parse_var(column.var)
-        var.update_type(column.type)
-        if not column.name.endswith("d"):
-          self.create_conditions(conditions, table, column.name, var)
-          
+        table = Table(self, "p%s" % self.translator.next_table_id, sub.name)
+        table.subquery = sub
+        self.translator.next_table_id += 1
+        if sub.optional:
+          table.join = "LEFT JOIN"
+          conditions = table.join_conditions
+        else:
+          conditions = self.conditions
           
+        for column in sub.columns:  
+          var = self.parse_var(column.var)
+          var.update_type(column.type)
+          if not column.name.endswith("d"):
+            self.create_conditions(conditions, table, column.name, var)
+              
   def parse_triples(self, triples):
     if self.triples: raise ValueError("Cannot parse triples twice!")
     self.block = triples
@@ -692,7 +751,7 @@ class SQLQuery(FuncSupport):
     
     if self.raw_selects == "*":
       self.vars_needed_for_select = { self.parse_var(var_name) for var_name in self.block.get_ordered_vars() }         
-      
+
     for i, triple in enumerate(self.triples): # Pass 0: Simple union blocks
       if isinstance(triple, UnionBlock) and triple.simple_union_triples:
         self.triples[i:i+1] = triple.simple_union_triples
@@ -701,9 +760,10 @@ class SQLQuery(FuncSupport):
     
     for triple in list(self.triples): # Pass 1: Determine var type and prop type
       if isinstance(triple, (Bind, Filter, Block)): continue
-      s, p, o = triple
-      
       triple.local_table_type = triple.table_type
+      
+      if triple.optional: continue # Optional => cannot be used to restrict variable type
+      s, p, o = triple
       if s.name == "VAR":
         var = self.parse_var(s)
         var.update_type("objs")
@@ -748,7 +808,7 @@ class SQLQuery(FuncSupport):
         
         prelim_triples = self.extract_triples(self.triples, vars, triple)
         
-        remnant_triples.remove(triple)
+        if triple in remnant_triples: remnant_triples.remove(triple)
         remnant_triples.difference_update(prelim_triples)
         prelim = self.translator.get_recursive_preliminary_select(triple, fixed, fixed_var, prelim_triples)
         triple.local_table_type = prelim.name
@@ -799,6 +859,7 @@ class SQLQuery(FuncSupport):
           o = self.parse_var(o)
           o.bindings.insert(0, """IN (SELECT o FROM %s%s)""" % (triple.local_table_type, extra))
           
+    conditions = self.conditions
     for triple in self.triples: # Pass 6: create triples tables and conditions
       if   isinstance(triple, Bind):
         self.parse_bind(triple)
@@ -822,23 +883,67 @@ class SQLQuery(FuncSupport):
       if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
       
       if self.name == "main": select_name = ""
-      else:                   select_name = "%s_" % self.name
-      table = Table("%sq%s" % (select_name, self.next_table_id), triple.local_table_type)
+      else:                   select_name = "%s_" % (self.name or "")
+      table = Table(self, "%sq%s" % (select_name, self.translator.next_table_id), triple.local_table_type)
       if triple.optional:
         table.join = "LEFT JOIN"
         conditions = table.join_conditions
       else:
         conditions = self.conditions
-      self.next_table_id += 1
-      self.tables.append(table)
-      self.name_2_table[table.name] = table
-      
+      self.translator.next_table_id += 1
+
       if triple.consider_s: self.create_conditions(conditions, table, "s", s)
       if triple.consider_p: self.create_conditions(conditions, table, "p", p, triple.likelihood_p)
       if triple.consider_o: self.create_conditions(conditions, table, "o", o, triple.likelihood_o)
       
       if p.modifier == "+": conditions.append("%s.nb>0"  % table.name)
-
+      
+      
+    if isinstance(triples, SimpleTripleBlock):
+      for static in triples.static_valuess:
+        if isinstance(static, StaticBlock):
+          if self.raw_selects == "*":
+            static.vars = sorted(static.all_vars)
+          else:
+            used_vars = set(self.vars)
+            used_vars.update(var.name for var in self.vars_needed_for_select)
+            static.vars = sorted(var for var in static.all_vars if (var in used_vars))
+          var_2_columns = defaultdict(list)
+          for column in static.translator.main_query.columns:
+              var_2_columns[column.var].append(column)
+              
+          static.translator.main_query.raw_selects = static.vars
+          static.translator.main_query.columns = []
+          static.translator.main_query.finalize_columns()
+          static.translator.solution_modifier = [None, None, None, None, None]
+          q = static.translator.finalize()
+          static.valuess = list(q._execute_sql())
+          
+        if len(static.vars) == 1:
+          var = self.parse_var(static.vars[0])
+          sql, sql_type, sql_d, sql_d_type = self._to_sql(var)
+          if sql is None:
+            var.bindings.append("IN (%s)" % ",".join(str(value[0]) for value in static.valuess))
+            var.static = static.valuess
+            var.type = "objs"
+          else:
+            conditions.append("%s IN (%s)" % (sql, ",".join(str(value[0]) for value in static.valuess)))
+            #conditions.append("LIKELIHOOD(%s IN (%s), %s)" % (sql, ",".join(str(value) for value in static.valuess),
+            #                                                  0.35 + 0.1 * (1.0 - math.exp(-len(static.valuess) / 100.0)))) # Favor statics with few elements
+            
+        else:
+          prelim = SQLStaticValuesPreliminaryQuery(self.translator, "static%s" % (len(self.translator.preliminary_selects) + 1), static)
+          self.translator.preliminary_selects.append(prelim)
+          table = Table(self, prelim.name, prelim.name)
+          for i, var in enumerate(static.vars):
+            var = self.parse_var(var)
+            sql, sql_type, sql_d, sql_d_type = self._to_sql(var)
+            if sql is None:
+              var.bindings.append("%s.col%s_o" % (prelim.name, i + 1))
+            else:
+              conditions.append("%s=%s.col%s_o" % (sql, prelim.name, i + 1))
+              
+              
   def get_fix_levels(self, vars0, exclude_triple = None):
     vars0_names = { var.name for var in vars0 }
     fix_levels  = defaultdict(float)
@@ -892,13 +997,11 @@ class SQLQuery(FuncSupport):
     return fix_levels
       
   def extract_triples(self, triples, vars, except_triple = None):
-    #print("   EXTRACT", triples, vars, except_triple)
     var_names = { var.name for var in vars }
     while True:
       r = [triple for triple in triples if (not triple == except_triple) and isinstance(triple, (Triple, Filter, Bind)) and (not triple.var_names.isdisjoint(var_names))]
       var_names2 = { var_name for triple in r for var_name in triple.var_names }
       if var_names2 == var_names:
-        #print("   EXTRACTED", r)
         return r
       var_names = var_names2
       
@@ -908,14 +1011,15 @@ class SQLQuery(FuncSupport):
     else:
       sql, sql_type, sql_d, sql_d_type = self._to_sql(x)
       
-      if table.query: # If datatype is 0 (=auto), disable datatype conditions and replace it by IS NOT NULL
-        for column in table.query.columns:
+      if table.subquery: # If datatype is 0 (=auto), disable datatype conditions and replace it by IS NOT NULL
+        for column in table.subquery.columns:
           if column.name == "%sd" % n[:-1]:
             if str(column.binding) == "0":
               sql_d = None
               conditions.append("%s.%sd IS NOT NULL" % (table.name, n[:-1])) # Datatype part
             break
           
+          
       if not sql is None:
         if isinstance(sql, str) and sql.startswith("IN "): operator = " "; sqls = self.parse_var(x).bindings
         else:                                              operator = "="; sqls = [sql]
@@ -924,14 +1028,28 @@ class SQLQuery(FuncSupport):
           condition = "%s.%s%s%s"  % (table.name, n, operator, sql)
           if likelihood is None: conditions.append(condition)
           else:                  conditions.append("LIKELIHOOD(%s, %s)" % (condition, likelihood))
-      if not sql_d is None:
+      if (not sql_d is None) and (n != "s") and (n != "p") and (table.name != "objs"):
         conditions.append("%s.%sd=%s" % (table.name, n[:-1], sql_d)) # Datatype part
         
-      if x.name == "VAR": x = self.vars[x.value]
-      if   isinstance(x, Variable):
+      if isinstance(x, rply.Token) and (x.name == "VAR"): x = self.vars[x.value]
+      if isinstance(x, Variable):
         if not x.initial_query: x.initial_query = self
         x.bindings.append("%s.%s" % (table.name, n))
         
+  def create_in_conditions(self, conditions, x, prelim):
+    assert len(prelim.columns) == 1
+    if isinstance(x, rply.Token): assert x.name == "VAR"
+    column = prelim.columns[0]
+    var = self.parse_var(x)
+    sql, sql_type, sql_d, sql_d_type = self._to_sql(var)
+    if sql is None:
+      var.bindings.append("IN (SELECT %s FROM %s)" % (column.name, prelim.name))
+      var.update_type(column.type)
+      var.in_select = prelim
+    else:
+      conditions.append("%s IN (SELECT %s FROM %s)" % (sql, column.name, prelim.name))
+      
+      
   def is_fixed(self, x):
     if x.name != "VAR": return True
     x = self.parse_var(x)
@@ -942,16 +1060,20 @@ class SQLQuery(FuncSupport):
   def finalize_columns(self):
     selected_parameter_index = 0
     i = j = 0
+
     if   self.raw_selects == "*": selects = [self.vars[var] for var in self.block.get_ordered_vars() if not var.startswith("??")]      
     elif self.raw_selects:        selects = self.raw_selects
-    elif self.tables:             selects = ["%s.s" % self.tables[0].name]
+    elif self.tables:
+      if self.tables[0].type in { "objs", "datas", "quads" }:
+                                  selects = ["1"] # Nothing to select (see TestSPARQL.test_128)
+      else:                       selects = ["%s.%s" % (self.tables[0].name, col) for col in self.translator.table_type_2_cols[self.tables[0].type]]
     else:                         selects = []
     
     def do_select(select):
       nonlocal selected_parameter_index
       if isinstance(select, str) and select.startswith("?"): select = self.vars[select]
       sql, sql_type, sql_d, sql_d_type = self._to_sql(select)
-      
+
       if   isinstance(select, rply.Token) and (select.name == "VAR"): var_name = select.value
       elif isinstance(select, Variable):                              var_name = select.name
       else:                                                           var_name = None
@@ -981,8 +1103,19 @@ class SQLQuery(FuncSupport):
           sql_type = "objs"
         else:
           raise ValueError("Cannot select '%s'!" % select)
-      self.columns.append(Column(var_name, sql_type,   sql,   "col%s_o" % i, j)); j += 1
-      if not sql_d is None: self.columns.append(Column(var_name, sql_d_type, sql_d, "col%s_d" % i, j)); j += 1
+        
+      self.columns.append(Column(var_name, sql_type, sql, "col%s_o" % i, j)); j += 1
+      if not sql_d is None:
+        self.columns.append(Column(var_name, sql_d_type, sql_d, "col%s_d" % i, j)); j += 1
+        
+    if self.preliminary:
+      self.translator.table_type_2_cols[self.name] = [column.name for column in self.columns]
+    
+  def set_column_names(self, names):
+    for column, name in zip(self.columns, names): column.name = name
+    
+    if self.preliminary:
+      self.translator.table_type_2_cols[self.name] = names
       
   def _to_sql(self, x):
     if isinstance(x, rply.Token) and (x.name == "VAR"): x = self.parse_var(x)
@@ -1001,7 +1134,10 @@ class SQLQuery(FuncSupport):
           else:
             return binding, "datas", x.fixed_datatype, "datas"
         type = "datas" if x.type == "datas" else "quads"
-        return binding, type, "%sd" % binding[:-1], type
+        if binding.startswith("static"): # no 'd' for static yet
+          return binding, type, 0, type
+        else:
+          return binding, type, "%sd" % binding[:-1], type
     elif x.name == "IRI":   return x.storid, "objs", None, None
     elif x.name == "PARAM": return "?%s" % x.number, "objs", None, None # XXX data parameter
     else:
@@ -1031,6 +1167,7 @@ class SQLCompoundQuery(object):
     for i, query in enumerate(self.queries):
       if i != 0: sql += "\n%s\n" % query.operator
       sql += query.sql()
+      
     if self.preliminary: return """%s(%s) AS (%s)""" % (self.name, ", ".join(column.name for column in self.columns), sql)
     return sql
   
@@ -1050,7 +1187,7 @@ class SQLCompoundQuery(object):
       for i, column in enumerate(query.columns): # Re-index columns
         column.index = i
     self.columns = self.queries[0].columns
-    
+
     
 class SQLNestedQuery(SQLQuery):
   def __init__(self, name):
@@ -1120,18 +1257,18 @@ class SQLRecursivePreliminaryQuery(SQLQuery):
       inversed_ps = [i for i in p if     i.inversed]
       if direct_ps:
         if len(direct_ps) == 1:
-          self.create_conditions(p_direct_conditions, Table("q", "quads" if self.need_d else "objs"), "p", direct_ps[0])
+          self.create_conditions(p_direct_conditions, Table(None, "q", "quads" if self.need_d else "objs"), "p", direct_ps[0])
         else:
           p_direct_conditions  .append("q.p IN (%s)" % ",".join(str(self._to_sql(i)[0]) for i in direct_ps))
           
       if inversed_ps:
         if len(inversed_ps) == 1:
-          self.create_conditions(p_inversed_conditions, Table("q", "quads" if self.need_d else "objs"), "p", inversed_ps[0])
+          self.create_conditions(p_inversed_conditions, Table(None, "q", "quads" if self.need_d else "objs"), "p", inversed_ps[0])
         else:
           p_inversed_conditions.append("q.p IN (%s)" % ",".join(str(self._to_sql(i)[0]) for i in inversed_ps))
           
     else:
-      self.create_conditions(p_direct_conditions, Table("q", "quads" if self.need_d else "objs"), "p", p)
+      self.create_conditions(p_direct_conditions, Table(None, "q", "quads" if self.need_d else "objs"), "p", p)
       
     self.extra_sql = ""
     if p_direct_conditions:
@@ -1159,4 +1296,23 @@ SELECT q.%s%s%s%s FROM %s q, %s rec WHERE %s %sAND q.%s=rec.%s""" % (
   self.name, " AND ".join(p_inversed_conditions),
   "AND rec.nb=0 " if p.modifier == "?" else "",
   self.non_fixed, self.non_fixed)
-  
+
+    
+      
+class SQLStaticValuesPreliminaryQuery(object):
+  recursive = False
+  def __init__(self, translator, name, static_values):
+    self.translator    = translator
+    self.name          = name
+    self.static_values = static_values
+    self.translator.table_type_2_cols[self.name] = ["col%s_o" % (i+1) for i in range(len(self.static_values.valuess[0]))]
+    
+    self.columns = []
+    for i in range(len(self.static_values.valuess[0])):
+      self.columns.append(Column(static_values.vars[i], "objs", "xxx", "col%s_o" % (i+1), i))
+    
+  def sql(self):
+    return "%s(%s) AS (VALUES %s)" % (self.name,
+                                      ",".join("col%s_o" % (i+1) for i in range(len(self.static_values.valuess[0]))),
+                                      ",".join("(%s)" % ",".join(str(value) for value in values) for values in self.static_values.valuess) )
+
